@@ -19,6 +19,7 @@
 import weakref
 import gobject
 import logging
+import imghdr
 
 import dbus
 import telepathy
@@ -27,8 +28,7 @@ import pymsn
 import contacts
 import conversation
 
-
-logger = logging.getLogger('telepathy-butterfly:conn')
+logger = logging.getLogger('telepathy-butterfly:connection')
 
 MSN_INTERFACE_DEBUG = "org.freedesktop.TpMSN.Debug"
 
@@ -72,10 +72,12 @@ MsnPassportHandle = telepathy.server.Handle
 
 MsnContactListHandle = telepathy.server.Handle
 
-class MsnConnection(telepathy.server.Connection, \
-        pymsn.Client, \
-        telepathy.server.ConnectionInterfacePresence, \
-        telepathy.server.ConnectionInterfaceAliasing):
+
+class MsnConnection(telepathy.server.Connection, 
+                    pymsn.Client, 
+                    telepathy.server.ConnectionInterfacePresence, 
+                    telepathy.server.ConnectionInterfaceAliasing, 
+                    telepathy.server.ConnectionInterfaceAvatars):
 
     _mandatory_parameters = {   'account':'s',
                                 'password':'s'
@@ -117,14 +119,16 @@ class MsnConnection(telepathy.server.Connection, \
         pymsn.Client.__init__(self, server, account, proxies = proxies)
         telepathy.server.ConnectionInterfacePresence.__init__(self)
         telepathy.server.ConnectionInterfaceAliasing.__init__(self)
+        telepathy.server.ConnectionInterfaceAvatars.__init__(self)
         
         self.profile.connect("notify::presence", self._user_presence_update_cb)
         self.profile.connect("notify::friendly-name", self._user_friendly_name_update_cb)
-        self.profile.connect("notify::personal-message", self._user_presence_update_cb)
+        self.profile.connect("notify::personal-message", self._user_personal_message_update_cb)
 
         self._die = False
         self._manager = manager
-        
+
+        self._avatars = {}
         self._contacts_handles = weakref.WeakValueDictionary()
         self._list_handles = weakref.WeakValueDictionary()
         
@@ -139,23 +143,27 @@ class MsnConnection(telepathy.server.Connection, \
     # pymsn callbacks
     def on_connect_failure(self, proto):
         pymsn.Client.on_connect_failure(self, proto)
-        self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED, telepathy.CONNECTION_STATUS_REASON_NETWORK_ERROR)
+        self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED,
+                telepathy.CONNECTION_STATUS_REASON_NETWORK_ERROR)
 
     def on_login_failure(self, proto):
         pymsn.Client.on_login_failure(self, proto)
-        self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED, telepathy.CONNECTION_STATUS_REASON_AUTHENTICATION_FAILED)
+        self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED,
+                telepathy.CONNECTION_STATUS_REASON_AUTHENTICATION_FAILED)
 
     def on_login_success(self, proto):
-        self.StatusChanged(telepathy.CONNECTION_STATUS_CONNECTED, telepathy.CONNECTION_STATUS_REASON_REQUESTED)
+        self.StatusChanged(telepathy.CONNECTION_STATUS_CONNECTED,
+                telepathy.CONNECTION_STATUS_REASON_REQUESTED)
         pymsn.Client.on_login_success(self, proto)
 
     def on_disconnected(self, conn):
         pymsn.Client.on_disconnected(self, conn)
         if self._die:
-            self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED, telepathy.CONNECTION_STATUS_REASON_REQUESTED)
-            gobject.idle_add(self._manager.disconnected, self)
+            reason = telepathy.CONNECTION_STATUS_REASON_REQUESTED
         else:
-            gobject.idle_add(self._connect_cb)
+            reason = telepathy.CONNECTION_STATUS_REASON_NONE_SPECIFIED
+        self.StatusChanged(telepathy.CONNECTION_STATUS_DISCONNECTED, reason)
+        gobject.idle_add(self._manager.disconnected, self)
     
     def __split_cmd(self, cmd):
         name = cmd.name.decode("utf-8")
@@ -186,6 +194,11 @@ class MsnConnection(telepathy.server.Connection, \
             subscribe_channel = contacts.MsnSubscribeListChannel(self, subscribe_handle)
             self._list_channels[subscribe_handle] = subscribe_channel
             self.add_channel(subscribe_channel, subscribe_handle, suppress_handler=False)
+            # create channel for handling publishing (Forward List)
+            publish_handle = self._get_handle_for_list('publish')
+            publish_channel = contacts.MsnPublishListChannel(self, publish_handle)
+            self._list_channels[publish_handle] = publish_channel
+            self.add_channel(publish_channel, publish_handle, suppress_handler=False)
 
     def on_switchboard_invitation(self, proto, server, key, session, passport, friendly_name):
         sender = self._get_handle_for_contact(passport)
@@ -213,7 +226,8 @@ class MsnConnection(telepathy.server.Connection, \
         contact.connect("notify::presence", self._contact_presence_update_cb)
         contact.connect("notify::friendly-name", self._contact_friendly_name_update_cb)
         contact.connect("notify::personal-message", self._contact_presence_update_cb)
-
+        # For avatar update
+        contact.connect("notify::msnobject", self._contact_msnobject_update_cb)
         handle = self._get_handle_for_contact(passport)
 
         for chan in self._list_channels.values():
@@ -237,7 +251,7 @@ class MsnConnection(telepathy.server.Connection, \
         elif handle_type == telepathy.CONNECTION_HANDLE_TYPE_LIST:
             get_handle = self._get_handle_for_list
         else:
-            raise NotAvailable('Msn only support lists and contacts')
+            raise telepathy.NotAvailable('Msn only support lists and contacts')
         
         for name in names:
             handle = get_handle(name)
@@ -255,14 +269,14 @@ class MsnConnection(telepathy.server.Connection, \
             handle = self._handles[(telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id)]
 
             if handle_type != telepathy.CONNECTION_HANDLE_TYPE_CONTACT:
-                raise InvalidHandle('only contact handles are valid for text channels at the moment')
+                raise telepathy.InvalidHandle('only contact handles are valid for text channels at the moment')
 
             if handle in self._im_channels:
                 chan = self._im_channels[handle]
             else:
                 invitee = self.get_contact_by_passport(handle.get_name())
                 if invitee.presence == pymsn.PresenceStatus.OFFLINE:
-                    raise NotAvailable('Contact not available')
+                    raise telepathy.NotAvailable('Contact not available')
                 chan = conversation.MsnTextConversation(self, [invitee])
                 self._im_channels[handle] = chan
         elif type == telepathy.CHANNEL_TYPE_CONTACT_LIST:
@@ -283,60 +297,38 @@ class MsnConnection(telepathy.server.Connection, \
 
         return chan._object_path
 
+    # Statuses
     def GetStatuses(self):
         # the arguments are in common to all on-line presences
-        arguments = {
-            'message':'s'
-        }
-
+        arguments = { 'message':'s' }
         # you get one of these for each status
         # {name:(type, self, exclusive, {argument:types}}
-        statuses = {
-            MsnPresence.ONLINE:(
-                telepathy.CONNECTION_PRESENCE_TYPE_AVAILABLE,
-                True, True, arguments
-            ),
-            MsnPresence.AWAY:(
-                telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
-                True, True, arguments
-            ),
-            MsnPresence.BUSY:(
-                telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
-                True, True, arguments
-            ),
-            MsnPresence.IDLE:(
-                telepathy.CONNECTION_PRESENCE_TYPE_EXTENDED_AWAY,
-                True, True, arguments
-            ),
-            MsnPresence.BRB:(
-                telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
-                True, True, arguments
-            ),
-            MsnPresence.PHONE:(
-                telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
-                True, True, arguments
-            ),
-            MsnPresence.LUNCH:(
-                telepathy.CONNECTION_PRESENCE_TYPE_EXTENDED_AWAY,
-                True, True, arguments
-            ),
-            MsnPresence.INVISIBLE:(
-                telepathy.CONNECTION_PRESENCE_TYPE_OFFLINE,
-                True, True, {}
-            ),
-            MsnPresence.OFFLINE:(
-                telepathy.CONNECTION_PRESENCE_TYPE_OFFLINE,
-                True, True, {}
-            )
+        return {
+            MsnPresence.ONLINE:( telepathy.CONNECTION_PRESENCE_TYPE_AVAILABLE,
+                True, True, arguments ),
+            MsnPresence.AWAY:( telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
+                True, True, arguments ),
+            MsnPresence.BUSY:( telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
+                True, True, arguments ),
+            MsnPresence.IDLE:( telepathy.CONNECTION_PRESENCE_TYPE_EXTENDED_AWAY,
+                True, True, arguments ),
+            MsnPresence.BRB:( telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
+                True, True, arguments ),
+            MsnPresence.PHONE:( telepathy.CONNECTION_PRESENCE_TYPE_AWAY,
+                True, True, arguments ),
+            MsnPresence.LUNCH:( telepathy.CONNECTION_PRESENCE_TYPE_EXTENDED_AWAY,
+                True, True, arguments ),
+            MsnPresence.INVISIBLE:( telepathy.CONNECTION_PRESENCE_TYPE_OFFLINE,
+                True, True, {} ),
+            MsnPresence.OFFLINE:( telepathy.CONNECTION_PRESENCE_TYPE_OFFLINE,
+                True, True, {} )
         }
 
-        return statuses
-
     def RequestPresence(self, contacts):
         presences = {}
         for handle_id in contacts:
-            self.check_handle(telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id)
-            handle = self._handles[telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id]
+            handle = self._get_handle_for_handle_id(
+                    telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id)
             passport = handle.get_name()
 
             contact = self.get_contact_by_passport(passport)
@@ -352,34 +344,82 @@ class MsnConnection(telepathy.server.Connection, \
         self.PresenceUpdate(presences)
 
     def SetStatus(self, statuses):
-        self.check_connected()
         status, arguments = statuses.items()[0]
         if status == MsnPresence.OFFLINE:
             self._die = True
 
         presence = MsnPresence.telepathy_to_pymsn[ status ]
         message = arguments['message']
-        self.personal_message = message.decode('UTF-8')
-        self.presence = presence
 
+        logger.debug("SetStatus: presence=%s, message=%s" % (presence, message))
+        if self._status != telepathy.CONNECTION_STATUS_CONNECTED:
+            self._initial_status = presence #TODO: initial personal mesage
+        else:
+            self.profile.personal_message = message.decode('UTF-8')
+            self.profile.presence = presence
+    
+    # Aliases
     def RequestAliases(self, contacts):
         result = []
         for handle_id in contacts:
-            if handle_id == self._self_handle:
+            handle = self._get_handle_for_handle_id(
+                    telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id)
+            if handle == self.GetSelfHandle():
                 result.append(unicode( self.profile.get_property("friendly-name"), 'UTF-8' ))
             else:
-                handle = self._handles[telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id]
                 contact = self.get_contact_by_passport(handle.get_name())
                 result.append(unicode(contact.get_property("friendly-name"), 'UTF-8'))
         return result
             
     def SetAliases(self, aliases):
-        for handle, alias in aliases.iteritems():
+        for handle_id, alias in aliases.iteritems():
+            handle = self._get_handle_for_handle_id(
+                    telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id)
             if handle != self.GetSelfHandle():
-                raise telepathy.PermissionDenied('MSN doesn\'t allow setting aliases for contacts')
-            self.friendly_name = alias
+                raise telepathy.PermissionDenied("MSN doesn't allow setting aliases for contacts")
+            self.profile.friendly_name = alias.decode('UTF-8')
+
+    # Avatar
+    def GetAvatarRequirements(self):
+        return (('image/jpeg', 'image/png', 'image/gif'),
+                    96, 96, 192, 192, 500 * 1024) #500kb
+
+    def GetAvatarTokens(self, contacts):
+        logger.debug("GetAvatarTokens() - contacts: %s" % contacts)
+        result = []
+        for handle_id in contacts:
+            handle = self._get_handle_for_handle_id(
+                    telepathy.CONNECTION_HANDLE_TYPE_CONTACT, handle_id)
+            if handle == self.GetSelfHandle():
+                result.append(self.profile.msnobject)
+            else:
+                contact = self.get_contact_by_passport(handle.get_name())
+                result.append(contact.get_property("msnobject"))
+
+    def RequestAvatar(self, handle_id):
+        logger.debug("RequestAvatar() - handle_id: %s" % handle_id)
+        avatar, mime_type = self._avatars.get(handle_id, ("",""))
+        if avatar:
+            logger.debug("RequestAvatar returns: (%s byte, %s) " % (len(avatar), mime_type))
+            return (avatar, mime_type)
+        logger.debug("RequestAvatar returns: None")
+
+    def SetAvatar(self, avatar, mime_type):
+        logger.debug("SetAvatar()")
+        if isinstance(avatar, str): # we got a dbus.ByteArray
+            self.profile.display_picture = avatar
+        else:
+            self.profile.display_picture = "".join([chr(b) for b in avatar])
+
+    def ClearAvatar():
+        logger.debug("ClearAvatar()")
+        self.profile.display_picture = None
 
     # Private methods
+    def _get_handle_for_handle_id(self, handle_type, handle_id):
+        self.check_handle(handle_type, handle_id)
+        return self._handles[handle_type, handle_id]
+
     def _get_handle_for_contact(self, passport):
         if passport in self._contacts_handles:
             handle = self._contacts_handles[passport]
@@ -415,7 +455,8 @@ class MsnConnection(telepathy.server.Connection, \
         """
         Called in the glib mainloop first iteration, to start connection
         """
-        self.StatusChanged(telepathy.CONNECTION_STATUS_CONNECTING, telepathy.CONNECTION_STATUS_REASON_REQUESTED)
+        self.StatusChanged(telepathy.CONNECTION_STATUS_CONNECTING,
+                telepathy.CONNECTION_STATUS_REASON_REQUESTED)
         self.login()
         return False
     
@@ -437,7 +478,29 @@ class MsnConnection(telepathy.server.Connection, \
             arguments['message'] = message
             
         self.PresenceUpdate({handle: (0, {presence:arguments})}) # :TODO: Timestamp ?
-    
+
+    def _contact_msnobject_update_cb(self, contact, property):
+        logger.debug("_contact_msnobject_update_cb() contact=%s property=%s" % (contact, property))
+        passport = contact.get_property("passport")
+        handle = self._get_handle_for_contact(passport)
+        avatar_token = contact.get_property('msnobject')
+        logger.debug("Avatar token: %s handle_id=%s" % (avatar_token, handle.get_id()))
+
+        def __on_dp_request_done(result):
+            if result == None:
+                logger.debug("_on_dp_request_done: Failed to fetch DP")
+                return
+        
+            logger.debug("_on_dp_request_done: Got DP for contact %s! %d bytes of data" % (contact, len(result)))
+            img_type = imghdr.what('', result)
+            if img_type is None: img_type = 'jpeg'
+            self._avatars[handle.get_id()] = (result, 'image/' + img_type)
+            self.AvatarUpdated(handle.get_id(), avatar_token)
+
+        # start fetching data
+        if contact.request_display_picture(__on_dp_request_done):
+            logger.debug("Requested avatar for contact: %s" % contact)
+
     def _user_friendly_name_update_cb(self, profile, property):
         passport = profile.get_property("passport")
         handle = self._get_handle_for_contact(passport)
@@ -457,6 +520,12 @@ class MsnConnection(telepathy.server.Connection, \
             
         self.PresenceUpdate({handle: (0, {presence:arguments})}) # :TODO: Timestamp ?
     
+    def _user_personal_message_update_cb(self, profile, property):
+        passport = profile.get_property("passport")
+        handle = self._get_handle_for_contact(passport)
+        friendly_name = unicode( profile.get_property("personal-message"), 'UTF-8' )
+        self.AliasesChanged(((handle,friendly_name),))
+
     # debug
     @dbus.service.signal(MSN_INTERFACE_DEBUG, signature='sss')
     def ProtocolIncoming(self, cmd, args, payload):
